import os
import pickle
import time
from copy import copy, deepcopy

import matplotlib
import sys
import torch
import numpy as np
from torch.hub import download_url_to_file
import matplotlib.pyplot as plt
import torch.utils.data
import torch.nn.functional as F
import json
import re

from tqdm import tqdm

plt.rcParams["figure.figsize"] = (7, 20) # size of window
plt.style.use('dark_background')

MIN_SENTENCE_LENGTH = 4
MAX_SENTENCE_LENGTH = 60
MIN_WORD_COUNT = 3

path_dataset = '../data/quotes.json'
if not os.path.exists(path_dataset):
    os.makedirs('../data', exist_ok=True)
    download_url_to_file(
        'http://share.yellowrobot.xyz/1630528570-intro-course-2021-q4/quotes.json',
        path_dataset,
        progress=True
    )

# 48000 samples
with open(f'{path_dataset}', 'r', encoding="utf8") as fp:
    qoutes_json =json.load(fp)

# TODO: Author, Category, Popularity

vocabulary = {}
authors = {}
categories = {}
sentences_lengths = {}

quotes_pre_processed = []


for quote in tqdm(qoutes_json, desc='quotes_pre_processed'):
    sentence = quote['Quote']
    sentence = sentence.lower()

    sentence = re.sub(r'\(.+\)', '', sentence)
    sentence = re.sub(r'\".+\"', '', sentence)
    sentence = re.sub(r'\{.+\}', '', sentence)
    sentence = re.sub(r'[^a-zA-Z\-\,\.\!\?\:\;\ ]', '', sentence)

    for it in ['~']:
        sentence = sentence.replace(it, '')

    for it in ['.', ',', '!', '?', ':', ';']:
        sentence = sentence.replace(it, ' ' + it)

    sentence_parts = sentence.split(' ')

    if len(sentence_parts) < MIN_SENTENCE_LENGTH:
        continue

    sentence_words = []
    is_invalid_detected = False
    for word in sentence_parts:
        word = word.strip()

        if len(word) == 0:
            continue

        is_invalid = False
        for it in ['{', '(']:
            if it in word:
                is_invalid = True
                is_invalid_detected = True
                break
        if is_invalid:
            continue

        if word not in vocabulary:
            vocabulary[word] = 0
        vocabulary[word] += 1
        sentence_words.append(word)

    if is_invalid_detected:
        print('invalid:', ' '.join(sentence_words))
        print('real: ', sentence)

    author = quote['Author'].lower().strip()
    if author not in authors:
        authors[author] = 0
    authors[author] += 1

    category = quote['Category'].lower().strip()
    if category not in categories:
        categories[category] = 0
    categories[category] += 1

    sentences_length = len(sentence_words)
    if sentences_length not in sentences_lengths:
        sentences_lengths[sentences_length] = 0
    sentences_lengths[sentences_length] += 1

    if len(sentence_words) < MIN_SENTENCE_LENGTH or len(sentence_words) > MAX_SENTENCE_LENGTH:
        continue

    quotes_pre_processed.append(
        (sentence_words, author, category)
    )


print('quotes_pre_processed', len(quotes_pre_processed))

# is_removed_word_any = True
# while is_removed_word_any:
#     is_removed_word_any = False
#
#     vocabulary_remmoved = []
#     for word, count in vocabulary.items():
#         if count < MIN_WORD_COUNT:
#             vocabulary_remmoved.append(word)
#     vocabulary_remmoved = set(vocabulary_remmoved)
#
#     vocabulary = {}
#     authors = {}
#     categories = {}
#     sentences_lengths = {}
#     quotes_pre_processed_before = [it for it in quotes_pre_processed]
#     quotes_pre_processed = []
#
#     quotes_set = set()
#
#     for sentence_words, author, category in tqdm(quotes_pre_processed_before, desc="final"):
#
#         hash_sent = ''.join(sentence_words)
#         if hash_sent in quotes_set:
#             is_removed_word_any = True
#             continue
#
#         quotes_set.add(hash_sent)
#
#         is_removed_word = False
#         for word in sentence_words:
#             if word in vocabulary_remmoved:
#                 is_removed_word = True
#                 is_removed_word_any = True
#                 break
#
#             if word not in vocabulary:
#                 vocabulary[word] = 0
#             vocabulary[word] += 1
#
#         if is_removed_word:
#             continue
#
#         if author not in authors:
#             authors[author] = 0
#         authors[author] += 1
#
#         if category not in categories:
#             categories[category] = 0
#         categories[category] += 1
#
#         sentences_length = len(sentence_words)
#         if sentences_length not in sentences_lengths:
#             sentences_lengths[sentences_length] = 0
#         sentences_lengths[sentences_length] += 1
#
#         quotes_pre_processed.append(
#             (sentence_words, author, category)
#         )
#
#
#     print('quotes_pre_processed', len(quotes_pre_processed))
#     print('vocabulary', len(vocabulary))
#     print('authors', len(authors))
#     print('categories', len(categories))

is_dupliate_quote_any = True
while is_dupliate_quote_any:
    is_dupliate_quote_any = False

    vocabulary_remmoved = []
    for word, count in vocabulary.items():
        if count < MIN_WORD_COUNT:
            vocabulary_remmoved.append(word)
    vocabulary_remmoved = set(vocabulary_remmoved)

    vocabulary = {}
    authors = {}
    categories = {}
    sentences_lengths = {}
    quotes_pre_processed_before = [it for it in quotes_pre_processed]
    quotes_pre_processed = []

    quotes_set = set()

    for sentence_words, author, category in tqdm(quotes_pre_processed_before, desc="final"):

        hash_sent = ''.join(sentence_words)
        if hash_sent in quotes_set:
            is_dupliate_quote_any = True
            continue

        quotes_set.add(hash_sent)

        for idx_w in range(len(sentence_words)):
            if sentence_words[idx_w] in vocabulary_remmoved:
                sentence_words[idx_w] = '[any]'
            if sentence_words[idx_w] not in vocabulary:
                vocabulary[sentence_words[idx_w]] = 0
            vocabulary[sentence_words[idx_w]] += 1

        if author not in authors:
            authors[author] = 0
        authors[author] += 1

        if category not in categories:
            categories[category] = 0
        categories[category] += 1

        sentences_length = len(sentence_words)
        if sentences_length not in sentences_lengths:
            sentences_lengths[sentences_length] = 0
        sentences_lengths[sentences_length] += 1

        quotes_pre_processed.append(
            (sentence_words, author, category)
        )


    print('quotes_pre_processed', len(quotes_pre_processed))
    print('vocabulary', len(vocabulary))
    print('authors', len(authors))
    print('categories', len(categories))
a  =0

# add eos token
vocabulary['[eos]'] = 0
for sentence_words, author, category in tqdm(quotes_pre_processed, desc='eos'):
    sentence_words.append('[eos]')
    vocabulary['[eos]'] += 1

sorted_vocabulary = sorted(list(vocabulary.items()), key=lambda it: it[1], reverse=True)
vocabulary_keys = [it[0] for it in sorted_vocabulary]
vocabulary_counts = [it[1] for it in sorted_vocabulary]

sorted_categories = sorted(list(categories.items()), key=lambda it: it[1], reverse=True)
categories_keys = [it[0] for it in sorted_categories]

sorted_authors = sorted(list(authors.items()), key=lambda it: it[1], reverse=True)
authors_keys = [it[0] for it in sorted_authors]

final_quotes_sentences = []
final_authors = []
final_categories = []
for sentence_words, author, category in tqdm(quotes_pre_processed, desc="final"):
    print(' '.join(sentence_words))
    final_quotes_sentences.append( [vocabulary_keys.index(it) for it in sentence_words] )
    final_authors.append(authors_keys.index(author))
    final_categories.append(categories_keys.index(category))

with open('../data/quotes.pkl', 'wb') as fp:
    pickle.dump((
        final_quotes_sentences, final_authors, final_categories,
        vocabulary_keys, vocabulary_counts, authors_keys, categories_keys
    ), fp)

max_len = 100
idxes = np.arange(max_len)
plt.barh(idxes, [it[1] for it in  sorted_vocabulary[:max_len]])
plt.yticks(idxes, [it[0] for it in  sorted_vocabulary[:max_len]])
plt.show()

max_len = 100
idxes = np.arange(max_len)
plt.barh(idxes, [it[1] for it in  sorted_vocabulary[-max_len:]])
plt.yticks(idxes, [it[0] for it in  sorted_vocabulary[-max_len:]])
plt.show()



max_len = len(sorted_categories)
idxes = np.arange(max_len)
plt.title('categories')
plt.barh(idxes, [it[1] for it in sorted_categories[:max_len]])
plt.yticks(idxes, [it[0] for it in sorted_categories[:max_len]])
plt.show()


max_len = 100
idxes = np.arange(max_len)
plt.title('authors')
plt.barh(idxes, [it[1] for it in  sorted_authors[:max_len]])
plt.yticks(idxes, [it[0] for it in  sorted_authors[:max_len]])
plt.show()

sorted_sentences_lengths = sorted(list(sentences_lengths.items()), key=lambda it: it[0], reverse=True)
max_len = len(sorted_sentences_lengths)
idxes = np.arange(max_len)
plt.title('sentence lengths')
plt.barh(idxes, [it[1] for it in  sorted_sentences_lengths[:max_len]])
plt.yticks(idxes, [it[0] for it in  sorted_sentences_lengths[:max_len]])
plt.show()

print(False)